# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'mslib/msui/ui/ui_airdata_dockwidget.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_AirdataDockwidget(object):
    def setupUi(self, AirdataDockwidget):
        AirdataDockwidget.setObjectName("AirdataDockwidget")
        AirdataDockwidget.resize(953, 106)
        self.verticalLayout = QtWidgets.QVBoxLayout(AirdataDockwidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.cbDrawAirports = QtWidgets.QCheckBox(AirdataDockwidget)
        self.cbDrawAirports.setMinimumSize(QtCore.QSize(145, 0))
        self.cbDrawAirports.setObjectName("cbDrawAirports")
        self.horizontalLayout_6.addWidget(self.cbDrawAirports)
        self.cbAirportType = CheckableComboBox(AirdataDockwidget)
        self.cbAirportType.setMinimumSize(QtCore.QSize(297, 0))
        self.cbAirportType.setLayoutDirection(QtCore.Qt.LeftToRight)
        self.cbAirportType.setObjectName("cbAirportType")
        self.horizontalLayout_6.addWidget(self.cbAirportType)
        self.btDownload = QtWidgets.QPushButton(AirdataDockwidget)
        self.btDownload.setMinimumSize(QtCore.QSize(135, 0))
        self.btDownload.setObjectName("btDownload")
        self.horizontalLayout_6.addWidget(self.btDownload)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_6.addItem(spacerItem)
        self.btApply = QtWidgets.QPushButton(AirdataDockwidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.btApply.setFont(font)
        self.btApply.setObjectName("btApply")
        self.horizontalLayout_6.addWidget(self.btApply)
        self.verticalLayout.addLayout(self.horizontalLayout_6)
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.cbDrawAirspaces = QtWidgets.QCheckBox(AirdataDockwidget)
        self.cbDrawAirspaces.setMinimumSize(QtCore.QSize(145, 0))
        self.cbDrawAirspaces.setObjectName("cbDrawAirspaces")
        self.horizontalLayout_7.addWidget(self.cbDrawAirspaces)
        self.cbAirspaces = CheckableComboBox(AirdataDockwidget)
        self.cbAirspaces.setMinimumSize(QtCore.QSize(297, 0))
        self.cbAirspaces.setObjectName("cbAirspaces")
        self.horizontalLayout_7.addWidget(self.cbAirspaces)
        self.btDownloadAsp = QtWidgets.QPushButton(AirdataDockwidget)
        self.btDownloadAsp.setMinimumSize(QtCore.QSize(135, 0))
        self.btDownloadAsp.setObjectName("btDownloadAsp")
        self.horizontalLayout_7.addWidget(self.btDownloadAsp)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_7.addItem(spacerItem1)
        self.verticalLayout.addLayout(self.horizontalLayout_7)
        self.horizontalLayout_8 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_8.setObjectName("horizontalLayout_8")
        self.cbFilterAirspaces = QtWidgets.QCheckBox(AirdataDockwidget)
        self.cbFilterAirspaces.setMinimumSize(QtCore.QSize(145, 0))
        self.cbFilterAirspaces.setObjectName("cbFilterAirspaces")
        self.horizontalLayout_8.addWidget(self.cbFilterAirspaces)
        self.label_3 = QtWidgets.QLabel(AirdataDockwidget)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout_8.addWidget(self.label_3)
        self.sbFrom = QtWidgets.QDoubleSpinBox(AirdataDockwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sbFrom.sizePolicy().hasHeightForWidth())
        self.sbFrom.setSizePolicy(sizePolicy)
        self.sbFrom.setMaximumSize(QtCore.QSize(190, 16777215))
        self.sbFrom.setMaximum(100.0)
        self.sbFrom.setObjectName("sbFrom")
        self.horizontalLayout_8.addWidget(self.sbFrom)
        self.label_4 = QtWidgets.QLabel(AirdataDockwidget)
        self.label_4.setObjectName("label_4")
        self.horizontalLayout_8.addWidget(self.label_4)
        self.sbTo = QtWidgets.QDoubleSpinBox(AirdataDockwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sbTo.sizePolicy().hasHeightForWidth())
        self.sbTo.setSizePolicy(sizePolicy)
        self.sbTo.setMaximumSize(QtCore.QSize(190, 16777215))
        self.sbTo.setPrefix("")
        self.sbTo.setMaximum(100.0)
        self.sbTo.setProperty("value", 100.0)
        self.sbTo.setObjectName("sbTo")
        self.horizontalLayout_8.addWidget(self.sbTo)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_8.addItem(spacerItem2)
        self.verticalLayout.addLayout(self.horizontalLayout_8)

        self.retranslateUi(AirdataDockwidget)
        QtCore.QMetaObject.connectSlotsByName(AirdataDockwidget)

    def retranslateUi(self, AirdataDockwidget):
        _translate = QtCore.QCoreApplication.translate
        AirdataDockwidget.setWindowTitle(_translate("AirdataDockwidget", "Form"))
        self.cbDrawAirports.setText(_translate("AirdataDockwidget", "draw airports"))
        self.btDownload.setToolTip(_translate("AirdataDockwidget", "Force download or update airports"))
        self.btDownload.setText(_translate("AirdataDockwidget", "Update/Download"))
        self.btApply.setToolTip(_translate("AirdataDockwidget", "Redraw the map with the current settings"))
        self.btApply.setText(_translate("AirdataDockwidget", "Apply"))
        self.btApply.setShortcut(_translate("AirdataDockwidget", "Return"))
        self.cbDrawAirspaces.setText(_translate("AirdataDockwidget", "draw airspaces"))
        self.btDownloadAsp.setToolTip(_translate("AirdataDockwidget", "Force download or update airspaces"))
        self.btDownloadAsp.setText(_translate("AirdataDockwidget", "Update/Download"))
        self.cbFilterAirspaces.setText(_translate("AirdataDockwidget", "filter airspaces"))
        self.label_3.setText(_translate("AirdataDockwidget", "from"))
        self.sbFrom.setSuffix(_translate("AirdataDockwidget", " km"))
        self.label_4.setText(_translate("AirdataDockwidget", "to"))
        self.sbTo.setSuffix(_translate("AirdataDockwidget", " km"))
from mslib.msui.mss_qt import CheckableComboBox
