# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'mslib/msui/ui/ui_topview_window.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_TopViewWindow(object):
    def setupUi(self, TopViewWindow):
        TopViewWindow.setObjectName("TopViewWindow")
        TopViewWindow.resize(952, 782)
        self.centralwidget = QtWidgets.QWidget(TopViewWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.mpl = MplTopViewWidget(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.mpl.sizePolicy().hasHeightForWidth())
        self.mpl.setSizePolicy(sizePolicy)
        self.mpl.setMinimumSize(QtCore.QSize(100, 100))
        self.mpl.setObjectName("mpl")
        self.horizontalLayout_2.addWidget(self.mpl)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.btMapRedraw = QtWidgets.QPushButton(self.centralwidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.btMapRedraw.setFont(font)
        self.btMapRedraw.setFlat(False)
        self.btMapRedraw.setObjectName("btMapRedraw")
        self.horizontalLayout.addWidget(self.btMapRedraw)
        spacerItem = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.btSettings = QtWidgets.QPushButton(self.centralwidget)
        self.btSettings.setObjectName("btSettings")
        self.horizontalLayout.addWidget(self.btSettings)
        spacerItem1 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.btRoundtrip = QtWidgets.QPushButton(self.centralwidget)
        self.btRoundtrip.setObjectName("btRoundtrip")
        self.horizontalLayout.addWidget(self.btRoundtrip)
        spacerItem2 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem2)
        self.cbTools = QtWidgets.QComboBox(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cbTools.sizePolicy().hasHeightForWidth())
        self.cbTools.setSizePolicy(sizePolicy)
        self.cbTools.setObjectName("cbTools")
        self.cbTools.addItem("")
        self.cbTools.addItem("")
        self.horizontalLayout.addWidget(self.cbTools)
        spacerItem3 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem3)
        self.cbChangeMapSection = QtWidgets.QComboBox(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cbChangeMapSection.sizePolicy().hasHeightForWidth())
        self.cbChangeMapSection.setSizePolicy(sizePolicy)
        self.cbChangeMapSection.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
        self.cbChangeMapSection.setObjectName("cbChangeMapSection")
        self.cbChangeMapSection.addItem("")
        self.cbChangeMapSection.addItem("")
        self.cbChangeMapSection.addItem("")
        self.cbChangeMapSection.addItem("")
        self.cbChangeMapSection.addItem("")
        self.cbChangeMapSection.addItem("")
        self.cbChangeMapSection.addItem("")
        self.horizontalLayout.addWidget(self.cbChangeMapSection)
        spacerItem4 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem4)
        self.verticalLayout.addLayout(self.horizontalLayout)
        TopViewWindow.setCentralWidget(self.centralwidget)
        self.actionCloseWindow = QtWidgets.QAction(TopViewWindow)
        self.actionCloseWindow.setObjectName("actionCloseWindow")
        TopViewWindow.addAction(self.actionCloseWindow)

        self.retranslateUi(TopViewWindow)
        self.actionCloseWindow.triggered.connect(TopViewWindow.close)
        QtCore.QMetaObject.connectSlotsByName(TopViewWindow)

    def retranslateUi(self, TopViewWindow):
        _translate = QtCore.QCoreApplication.translate
        TopViewWindow.setWindowTitle(_translate("TopViewWindow", "Top View - Mission Support System"))
        self.btMapRedraw.setToolTip(_translate("TopViewWindow", "Redraw Map (R)"))
        self.btMapRedraw.setText(_translate("TopViewWindow", "&REDRAW"))
        self.btMapRedraw.setShortcut(_translate("TopViewWindow", "R"))
        self.btSettings.setText(_translate("TopViewWindow", "options"))
        self.btRoundtrip.setText(_translate("TopViewWindow", "make roundtrip"))
        self.cbTools.setItemText(0, _translate("TopViewWindow", "(select to open tool)"))
        self.cbTools.setItemText(1, _translate("TopViewWindow", "WMS"))
        self.cbChangeMapSection.setItemText(0, _translate("TopViewWindow", "to reset map select a region"))
        self.cbChangeMapSection.setItemText(1, _translate("TopViewWindow", "Spitsbergen, large"))
        self.cbChangeMapSection.setItemText(2, _translate("TopViewWindow", "Spitsbergen, local"))
        self.cbChangeMapSection.setItemText(3, _translate("TopViewWindow", "Europe (ste)"))
        self.cbChangeMapSection.setItemText(4, _translate("TopViewWindow", "Germany (ste)"))
        self.cbChangeMapSection.setItemText(5, _translate("TopViewWindow", "Europe (cyl)"))
        self.cbChangeMapSection.setItemText(6, _translate("TopViewWindow", "Germany (cyl)"))
        self.actionCloseWindow.setText(_translate("TopViewWindow", "CloseWindow"))
        self.actionCloseWindow.setShortcut(_translate("TopViewWindow", "Ctrl+W"))
from mslib.msui.mpl_qtwidget import MplTopViewWidget
