# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui_kmloverlay_dockwidget.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_KMLOverlayDockWidget(object):
    def setupUi(self, KMLOverlayDockWidget):
        KMLOverlayDockWidget.setObjectName("KMLOverlayDockWidget")
        KMLOverlayDockWidget.resize(545, 233)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(KMLOverlayDockWidget.sizePolicy().hasHeightForWidth())
        KMLOverlayDockWidget.setSizePolicy(sizePolicy)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(KMLOverlayDockWidget)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setSizeConstraint(QtWidgets.QLayout.SetMinimumSize)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.listWidget = QtWidgets.QListWidget(KMLOverlayDockWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.listWidget.sizePolicy().hasHeightForWidth())
        self.listWidget.setSizePolicy(sizePolicy)
        self.listWidget.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
        self.listWidget.setObjectName("listWidget")
        self.horizontalLayout.addWidget(self.listWidget)
        self.frame = QtWidgets.QFrame(KMLOverlayDockWidget)
        self.frame.setObjectName("frame")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.frame)
        self.verticalLayout_4.setContentsMargins(-1, 5, -1, -1)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.pushButton_color = QtWidgets.QPushButton(self.frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushButton_color.sizePolicy().hasHeightForWidth())
        self.pushButton_color.setSizePolicy(sizePolicy)
        self.pushButton_color.setAutoFillBackground(False)
        self.pushButton_color.setCheckable(False)
        self.pushButton_color.setObjectName("pushButton_color")
        self.verticalLayout_4.addWidget(self.pushButton_color)
        self.dsbx_linewidth = QtWidgets.QDoubleSpinBox(self.frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.dsbx_linewidth.sizePolicy().hasHeightForWidth())
        self.dsbx_linewidth.setSizePolicy(sizePolicy)
        self.dsbx_linewidth.setWrapping(True)
        self.dsbx_linewidth.setAlignment(QtCore.Qt.AlignCenter)
        self.dsbx_linewidth.setReadOnly(False)
        self.dsbx_linewidth.setCorrectionMode(QtWidgets.QAbstractSpinBox.CorrectToNearestValue)
        self.dsbx_linewidth.setKeyboardTracking(True)
        self.dsbx_linewidth.setProperty("showGroupSeparator", False)
        self.dsbx_linewidth.setMinimum(0.1)
        self.dsbx_linewidth.setMaximum(10.0)
        self.dsbx_linewidth.setSingleStep(1.0)
        self.dsbx_linewidth.setObjectName("dsbx_linewidth")
        self.verticalLayout_4.addWidget(self.dsbx_linewidth)
        self.horizontalLayout.addWidget(self.frame)
        self.horizontalLayout_3.addLayout(self.horizontalLayout)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setSizeConstraint(QtWidgets.QLayout.SetMinimumSize)
        self.verticalLayout.setObjectName("verticalLayout")
        self.pushButton_select_all = QtWidgets.QPushButton(KMLOverlayDockWidget)
        self.pushButton_select_all.setObjectName("pushButton_select_all")
        self.verticalLayout.addWidget(self.pushButton_select_all)
        self.btSelectFile = QtWidgets.QPushButton(KMLOverlayDockWidget)
        self.btSelectFile.setObjectName("btSelectFile")
        self.verticalLayout.addWidget(self.btSelectFile)
        self.pushButton_remove = QtWidgets.QPushButton(KMLOverlayDockWidget)
        self.pushButton_remove.setObjectName("pushButton_remove")
        self.verticalLayout.addWidget(self.pushButton_remove)
        self.pushButton_unselect_all = QtWidgets.QPushButton(KMLOverlayDockWidget)
        self.pushButton_unselect_all.setObjectName("pushButton_unselect_all")
        self.verticalLayout.addWidget(self.pushButton_unselect_all)
        self.horizontalLayout_3.addLayout(self.verticalLayout)
        self.verticalLayout_2.addLayout(self.horizontalLayout_3)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.verticalLayout_2.addLayout(self.horizontalLayout_2)
        self.pushButton_merge = QtWidgets.QPushButton(KMLOverlayDockWidget)
        self.pushButton_merge.setObjectName("pushButton_merge")
        self.verticalLayout_2.addWidget(self.pushButton_merge)
        self.labelStatusBar = QtWidgets.QLabel(KMLOverlayDockWidget)
        self.labelStatusBar.setObjectName("labelStatusBar")
        self.verticalLayout_2.addWidget(self.labelStatusBar)

        self.retranslateUi(KMLOverlayDockWidget)
        QtCore.QMetaObject.connectSlotsByName(KMLOverlayDockWidget)

    def retranslateUi(self, KMLOverlayDockWidget):
        _translate = QtCore.QCoreApplication.translate
        KMLOverlayDockWidget.setWindowTitle(_translate("KMLOverlayDockWidget", "KML Overlay"))
        self.listWidget.setToolTip(_translate("KMLOverlayDockWidget", "KML Files;  The icon shows color and height corresponds to linewidth"))
        self.pushButton_color.setToolTip(_translate("KMLOverlayDockWidget", "Click to change the colour of the selected KML file"))
        self.pushButton_color.setText(_translate("KMLOverlayDockWidget", " Change Color"))
        self.dsbx_linewidth.setToolTip(_translate("KMLOverlayDockWidget", "Edit Linewidth here! Supported values are from 0.1 to 10"))
        self.pushButton_select_all.setToolTip(_translate("KMLOverlayDockWidget", "Selects (Checks) All the Files "))
        self.pushButton_select_all.setText(_translate("KMLOverlayDockWidget", "Select All Files"))
        self.btSelectFile.setToolTip(_translate("KMLOverlayDockWidget", "Add multiple KML Files"))
        self.btSelectFile.setText(_translate("KMLOverlayDockWidget", "Add KML Files"))
        self.pushButton_remove.setToolTip(_translate("KMLOverlayDockWidget", "Remove Checked KML Files "))
        self.pushButton_remove.setText(_translate("KMLOverlayDockWidget", "Remove Files"))
        self.pushButton_unselect_all.setToolTip(_translate("KMLOverlayDockWidget", "Unselects (unchecks) all Files"))
        self.pushButton_unselect_all.setText(_translate("KMLOverlayDockWidget", "Unselect All Files"))
        self.pushButton_merge.setToolTip(_translate("KMLOverlayDockWidget", "Merge multiple KML Files into one"))
        self.pushButton_merge.setText(_translate("KMLOverlayDockWidget", "Merge and Export KML File"))
        self.labelStatusBar.setText(_translate("KMLOverlayDockWidget", "Status : "))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    KMLOverlayDockWidget = QtWidgets.QWidget()
    ui = Ui_KMLOverlayDockWidget()
    ui.setupUi(KMLOverlayDockWidget)
    KMLOverlayDockWidget.show()
    sys.exit(app.exec_())

