# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui_linearview_options.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_LinearViewOptionsDialog(object):
    def setupUi(self, LinearViewOptionsDialog):
        LinearViewOptionsDialog.setObjectName("LinearViewOptionsDialog")
        LinearViewOptionsDialog.resize(614, 180)
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(LinearViewOptionsDialog)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.groupBox_5 = QtWidgets.QGroupBox(LinearViewOptionsDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox_5.sizePolicy().hasHeightForWidth())
        self.groupBox_5.setSizePolicy(sizePolicy)
        self.groupBox_5.setObjectName("groupBox_5")
        self.gridLayout_4 = QtWidgets.QGridLayout(self.groupBox_5)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.lv_cbtitlesize = QtWidgets.QComboBox(self.groupBox_5)
        self.lv_cbtitlesize.setObjectName("lv_cbtitlesize")
        self.lv_cbtitlesize.addItem("")
        self.lv_cbtitlesize.addItem("")
        self.lv_cbtitlesize.addItem("")
        self.lv_cbtitlesize.addItem("")
        self.lv_cbtitlesize.addItem("")
        self.lv_cbtitlesize.addItem("")
        self.lv_cbtitlesize.addItem("")
        self.lv_cbtitlesize.addItem("")
        self.lv_cbtitlesize.addItem("")
        self.lv_cbtitlesize.addItem("")
        self.lv_cbtitlesize.addItem("")
        self.lv_cbtitlesize.addItem("")
        self.lv_cbtitlesize.addItem("")
        self.lv_cbtitlesize.addItem("")
        self.lv_cbtitlesize.addItem("")
        self.lv_cbtitlesize.addItem("")
        self.gridLayout_4.addWidget(self.lv_cbtitlesize, 3, 1, 1, 1)
        self.label_2 = QtWidgets.QLabel(self.groupBox_5)
        self.label_2.setObjectName("label_2")
        self.gridLayout_4.addWidget(self.label_2, 3, 2, 1, 1)
        self.lv_cbaxessize = QtWidgets.QComboBox(self.groupBox_5)
        self.lv_cbaxessize.setObjectName("lv_cbaxessize")
        self.lv_cbaxessize.addItem("")
        self.lv_cbaxessize.addItem("")
        self.lv_cbaxessize.addItem("")
        self.lv_cbaxessize.addItem("")
        self.lv_cbaxessize.addItem("")
        self.lv_cbaxessize.addItem("")
        self.lv_cbaxessize.addItem("")
        self.lv_cbaxessize.addItem("")
        self.lv_cbaxessize.addItem("")
        self.lv_cbaxessize.addItem("")
        self.lv_cbaxessize.addItem("")
        self.lv_cbaxessize.addItem("")
        self.lv_cbaxessize.addItem("")
        self.lv_cbaxessize.addItem("")
        self.lv_cbaxessize.addItem("")
        self.lv_cbaxessize.addItem("")
        self.gridLayout_4.addWidget(self.lv_cbaxessize, 3, 3, 1, 1)
        self.label = QtWidgets.QLabel(self.groupBox_5)
        self.label.setObjectName("label")
        self.gridLayout_4.addWidget(self.label, 3, 0, 1, 1)
        self.verticalLayout_4.addWidget(self.groupBox_5)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_4.addItem(spacerItem)
        self.buttonBox = QtWidgets.QDialogButtonBox(LinearViewOptionsDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout_4.addWidget(self.buttonBox)

        self.retranslateUi(LinearViewOptionsDialog)
        self.buttonBox.accepted.connect(LinearViewOptionsDialog.accept)
        self.buttonBox.rejected.connect(LinearViewOptionsDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(LinearViewOptionsDialog)

    def retranslateUi(self, LinearViewOptionsDialog):
        _translate = QtCore.QCoreApplication.translate
        LinearViewOptionsDialog.setWindowTitle(_translate("LinearViewOptionsDialog", "Linear View Options"))
        self.groupBox_5.setTitle(_translate("LinearViewOptionsDialog", "Plot Options"))
        self.lv_cbtitlesize.setItemText(0, _translate("LinearViewOptionsDialog", "default"))
        self.lv_cbtitlesize.setItemText(1, _translate("LinearViewOptionsDialog", "4"))
        self.lv_cbtitlesize.setItemText(2, _translate("LinearViewOptionsDialog", "6"))
        self.lv_cbtitlesize.setItemText(3, _translate("LinearViewOptionsDialog", "8"))
        self.lv_cbtitlesize.setItemText(4, _translate("LinearViewOptionsDialog", "10"))
        self.lv_cbtitlesize.setItemText(5, _translate("LinearViewOptionsDialog", "12"))
        self.lv_cbtitlesize.setItemText(6, _translate("LinearViewOptionsDialog", "14"))
        self.lv_cbtitlesize.setItemText(7, _translate("LinearViewOptionsDialog", "16"))
        self.lv_cbtitlesize.setItemText(8, _translate("LinearViewOptionsDialog", "18"))
        self.lv_cbtitlesize.setItemText(9, _translate("LinearViewOptionsDialog", "20"))
        self.lv_cbtitlesize.setItemText(10, _translate("LinearViewOptionsDialog", "22"))
        self.lv_cbtitlesize.setItemText(11, _translate("LinearViewOptionsDialog", "24"))
        self.lv_cbtitlesize.setItemText(12, _translate("LinearViewOptionsDialog", "26"))
        self.lv_cbtitlesize.setItemText(13, _translate("LinearViewOptionsDialog", "28"))
        self.lv_cbtitlesize.setItemText(14, _translate("LinearViewOptionsDialog", "30"))
        self.lv_cbtitlesize.setItemText(15, _translate("LinearViewOptionsDialog", "32"))
        self.label_2.setText(_translate("LinearViewOptionsDialog", " Axes Label Size  "))
        self.lv_cbaxessize.setItemText(0, _translate("LinearViewOptionsDialog", "default"))
        self.lv_cbaxessize.setItemText(1, _translate("LinearViewOptionsDialog", "4"))
        self.lv_cbaxessize.setItemText(2, _translate("LinearViewOptionsDialog", "6"))
        self.lv_cbaxessize.setItemText(3, _translate("LinearViewOptionsDialog", "8"))
        self.lv_cbaxessize.setItemText(4, _translate("LinearViewOptionsDialog", "10"))
        self.lv_cbaxessize.setItemText(5, _translate("LinearViewOptionsDialog", "12"))
        self.lv_cbaxessize.setItemText(6, _translate("LinearViewOptionsDialog", "14"))
        self.lv_cbaxessize.setItemText(7, _translate("LinearViewOptionsDialog", "16"))
        self.lv_cbaxessize.setItemText(8, _translate("LinearViewOptionsDialog", "18"))
        self.lv_cbaxessize.setItemText(9, _translate("LinearViewOptionsDialog", "20"))
        self.lv_cbaxessize.setItemText(10, _translate("LinearViewOptionsDialog", "22"))
        self.lv_cbaxessize.setItemText(11, _translate("LinearViewOptionsDialog", "24"))
        self.lv_cbaxessize.setItemText(12, _translate("LinearViewOptionsDialog", "26"))
        self.lv_cbaxessize.setItemText(13, _translate("LinearViewOptionsDialog", "28"))
        self.lv_cbaxessize.setItemText(14, _translate("LinearViewOptionsDialog", "30"))
        self.lv_cbaxessize.setItemText(15, _translate("LinearViewOptionsDialog", "32"))
        self.label.setText(_translate("LinearViewOptionsDialog", " Plot Title Size"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    LinearViewOptionsDialog = QtWidgets.QDialog()
    ui = Ui_LinearViewOptionsDialog()
    ui.setupUi(LinearViewOptionsDialog)
    LinearViewOptionsDialog.show()
    sys.exit(app.exec_())
